import random

from egse.ariel.tcu import TcuMode, NUM_TSM_PROBES_PER_FRAME, NUM_TSM_FRAMES, NUM_M2MD_AXES
from egse.ariel.tcu.tcu import TcuInterface, TcuSimulator
from egse.ariel.tcu.tcu_cmd_utils import (
    CommandAddress,
    GeneralCommandIdentifier,
    TSMCommandIdentifier,
    TCU_LOGICAL_ADDRESS,
    DATA_LENGTH,
    format_value,
    PacketType,
    M2MDCommandIdentifier,
    HKCommandIdentifier,
)
import logging

from egse.device import DeviceTransport
from egse.mixin import DynamicCommandMixin

LOGGER = logging.getLogger("egse.ariel.tcu.tcu")

EXPECTED_TRANSACTION_ID = -1

###############################
# Test building command strings
###############################


class TcuTestInterface(DeviceTransport):
    """Dummy device transport for testing."""

    def trans(self, command: str) -> bytes:
        """Returns the given command string as bytes.

        This is used to test whether the command string is built correctly (and the transaction identifier is
        incremented for each command call).

        Args:
            command (str): Command string to encode.

        Returns:
            Encoded command string.
        """

        global EXPECTED_TRANSACTION_ID
        EXPECTED_TRANSACTION_ID += 1

        return command.encode()


class TcuTest(TcuInterface, DynamicCommandMixin):
    """Dummy TCU interface for testing."""

    def __init__(self):
        super().__init__()
        self.transport = self.tcu = TcuTestInterface()


TCU_TEST = TcuTest()


TRANSACTION_ID_OFFSET = 5
DATA_LENGTH_OFFSET = 10
ADDRESS_OFFSET = 15
CMD_ID_OFFSET = 20
CARGO1_OFFSET = 25
CARGO2_OFFSET = 30


def is_tcu_read_cmd(cmd_string: str) -> bool:
    """Checks whether the given command is a TCU read command.

    Args:
        cmd_string (str): Command string to check.

    Returns:
        True of the given command is a TCU read command; False otherwise.
    """

    expected_prefix = f"{TCU_LOGICAL_ADDRESS}{PacketType.READ.value}"
    return cmd_string.startswith(expected_prefix)


def is_tcu_write_cmd(cmd_string: str) -> bool:
    """Checks whether the given command is a TCU write command.

    Args:
        cmd_string (str): Command string to check.

    Returns:
        True of the given command is a TCU write command; False otherwise.
    """

    expected_prefix = f"{TCU_LOGICAL_ADDRESS}{PacketType.WRITE.value}"
    return cmd_string.startswith(expected_prefix)


def get_transaction_id(cmd_string: str) -> str:
    """Extracts the transaction identifier from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the transaction identifier from.

    Returns:
        Transaction identifier.
    """

    transaction_id_hex_str = cmd_string[TRANSACTION_ID_OFFSET : TRANSACTION_ID_OFFSET + 4]

    return transaction_id_hex_str


def get_data_length(cmd_string: str) -> str:
    """Extracts the data length from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the data length from.

    Returns:
        Data length.
    """

    return cmd_string[DATA_LENGTH_OFFSET : DATA_LENGTH_OFFSET + 4]


def get_address(cmd_string: str) -> str:
    """Extracts the command address from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the command address from.

    Returns:
        Command address.
    """

    return cmd_string[ADDRESS_OFFSET : ADDRESS_OFFSET + 4]


def get_cmd_id(cmd_string: str) -> str:
    """Extracts the command identifier from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the command identifier from.

    Returns:
        Command identifier.
    """

    return cmd_string[CMD_ID_OFFSET : CMD_ID_OFFSET + 4]


def get_cargo1(cmd_string: str) -> str:
    """Extracts the cargo1 argument from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the cargo1 argument from.

    Returns:
        Cargo1 argument from the given TCU command string.
    """

    return cmd_string[CARGO1_OFFSET : CARGO1_OFFSET + 4]


def get_cargo2(cmd_string: str) -> str:
    """Extracts the cargo2 argument from the given TCU command string.

    Args:
        cmd_string (str): Command string to extract the cargo2 argument from.

    Returns:
        Cargo2 argument from the given TCU command string.
    """

    return cmd_string[CARGO2_OFFSET : CARGO2_OFFSET + 4]


def get_random_hex() -> int:
    """Generates a random hex string of maximum length 4, without leading "0x".

    Returns:
        Random hex string of maximum length 4, without leading "0x".
    """

    # FIXME

    return random.getrandbits(16)


def get_expected_transaction_id_as_hex() -> str:
    """Returns the expected transaction identifier.

    The leading "0x" is stripped off and the transaction identifier is padded with leading zeros to ensure it is
    always 4 characters long.

    Returns:
        Expected transaction identifier.
    """

    return f"{EXPECTED_TRANSACTION_ID:X}".zfill(4)


def test_tcu_firmware_id():
    cmd_string = TCU_TEST.tcu_firmware_id().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_FIRMWARE_ID.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0000 0000 0000 0000")


def test_get_tcu_mode():
    cmd_string = TCU_TEST.get_tcu_mode().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_MODE.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0000 0001 0000 0000")


def test_set_tcu_mode():
    for tcu_mode in TcuMode:
        cmd_string = TCU_TEST.set_tcu_mode(tcu_mode=tcu_mode).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.GENERAL.value
        assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_MODE.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(tcu_mode)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 0000 0001 0000 {format_value(tcu_mode)}"
        )

    for tcu_mode in TcuMode:
        cmd_string = TCU_TEST.set_tcu_mode(tcu_mode=tcu_mode.value).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.GENERAL.value
        assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_MODE.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(tcu_mode)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 0000 0001 0000 {format_value(tcu_mode)}"
        )


def test_tcu_status():
    cmd_string = TCU_TEST.tcu_status().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_STATUS.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0000 0002 0000 0000")


def test_tcu_simulated():
    simulated = get_random_hex()
    cmd_string = TCU_TEST.tcu_simulated(cargo2=simulated).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.TCU_SIMULATED.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == format_value(simulated)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0000 0003 0000 {format_value(simulated)}"
    )


def test_restart_links_period_latch():
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.restart_links_period_latch(cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.RESTART_LINKS_PERIOD_LATCH.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0000 0004 0000 {format_value(cargo2)}"
    )


def test_get_restart_links_period():
    cmd_string = TCU_TEST.get_restart_links_period().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.RESTART_LINKS_PERIOD.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0000 0005 0000 0000")


def test_set_restart_links_period():
    link_period = get_random_hex()
    cmd_string = TCU_TEST.set_restart_links_period(link_period=link_period).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.GENERAL.value
    assert get_cmd_id(cmd_string) == GeneralCommandIdentifier.RESTART_LINKS_PERIOD.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == format_value(link_period)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0000 0005 0000 {format_value(link_period)}"
    )


def test_ope_mng_command():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_command(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_COMMAND.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 0000 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_command(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_COMMAND.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 0000 0000 {format_value(cargo2)}"
        )


def test_ope_mng_event_clear_protect_flag():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_event_clear_protect_flag(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR_PROTECT_FLAG.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 0001 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_event_clear_protect_flag(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR_PROTECT_FLAG.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 0001 0000 {format_value(cargo2)}"
        )


def test_ope_mng_event_clear():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_event_clear(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 0002 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.ope_mng_event_clear(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_CLEAR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 0002 0000 {format_value(cargo2)}"
        )


def test_ope_mng_status():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.ope_mng_status(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_STATUS.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 0003 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.ope_mng_status(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_STATUS.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 0003 0000 0000"
        )


def test_ope_mng_event_reg():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.ope_mng_event_reg(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_REG.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 0004 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.ope_mng_event_reg(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.OPE_MNG_EVENT_REG.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 0004 0000 0000"
        )


def test_get_acq_curr_off_corr():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.get_acq_curr_off_corr(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_OFF_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1000 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.get_acq_curr_off_corr(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_OFF_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1000 0000 0000"
        )


def test_set_acq_curr_off_corr():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_acq_curr_off_corr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_OFF_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1000 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_acq_curr_off_corr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_OFF_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1000 0000 {format_value(cargo2)}"
        )


def test_get_acq_curr_gain_corr():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.get_acq_curr_gain_corr(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1001 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.get_acq_curr_gain_corr(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1001 0000 0000"
        )


def test_set_acq_curr_gain_corr():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_acq_curr_gain_corr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1001 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_acq_curr_gain_corr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_CURR_GAIN_CORR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1001 0000 {format_value(cargo2)}"
        )


def test_acq_axis_a_curr_read():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.acq_axis_a_curr_read(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AXIS_A_CURR_READ.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1002 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.acq_axis_a_curr_read(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AXIS_A_CURR_READ.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1002 0000 0000"
        )


def test_acq_axis_b_curr_read():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.acq_axis_b_curr_read(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AXIS_B_CURR_READ.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1003 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.acq_axis_b_curr_read(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AXIS_B_CURR_READ.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1003 0000 0000"
        )


def test_acq_ave_lpf_en():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_ave_lpf_en(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVE_LPF_EN.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1004 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_ave_lpf_en(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVE_LPF_EN.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1004 0000 {format_value(cargo2)}"
        )


def test_acq_ovc_cfg_filter():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_ovc_cfg_filter(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_OVC_CFG_FILTER.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1005 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_ovc_cfg_filter(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_OVC_CFG_FILTER.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1005 0000 {format_value(cargo2)}"
        )


# noinspection SpellCheckingInspection
def test_acq_avc_filt_time():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_avc_filt_time(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVC_FILT_TIME.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1006 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_avc_filt_time(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVC_FILT_TIME.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1006 0000 {format_value(cargo2)}"
        )


def test_acq_average_type():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_average_type(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVERAGE_TYPE.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1007 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_average_type(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_AVERAGE_TYPE.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1007 0000 {format_value(cargo2)}"
        )


# noinspection SpellCheckingInspection
def test_acq_spk_filt_counter_lim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_spk_filt_counter_lim(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_SPK_FILT_COUNTER_LIM.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1008 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_spk_filt_counter_lim(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_SPK_FILT_COUNTER_LIM.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1008 0000 {format_value(cargo2)}"
        )


# noinspection SpellCheckingInspection
def test_acq_spk_filt_incr_thr():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_spk_filt_incr_thr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_SPK_FILT_INCR_THR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 1009 0000 {format_value(cargo2)}"
        )

    for axis in range(1, 5):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.acq_spk_filt_incr_thr(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.ACQ_SPK_FILT_INCR_THR.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 1009 0000 {format_value(cargo2)}"
        )


def test_get_prof_gen_axis_step():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.get_prof_gen_axis_step(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STEP.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2000 0000 0000")

    for axis in range(1, 5):
        cmd_string = TCU_TEST.get_prof_gen_axis_step(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STEP.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2000 0000 0000"
        )


def test_set_prof_gen_axis_step():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_step(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STEP.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2000 0000 {format_value(cargo2)}"
        )

    for axis in range(1, 5):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_step(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STEP.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2000 0000 {format_value(cargo2)}"
        )


def test_get_prof_gen_axis_speed():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.get_prof_gen_axis_speed(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2001 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.get_prof_gen_axis_speed(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2001 0000 0000"
        )


def test_set_prof_gen_axis_speed():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_speed(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2001 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_speed(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_SPEED.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2001 0000 {format_value(cargo2)}"
        )


def test_get_prof_gen_axis_state_start():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cmd_string = TCU_TEST.get_prof_gen_axis_state_start(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2002 0000 0000")

    for axis in range(1, NUM_M2MD_AXES + 1):
        cmd_string = TCU_TEST.get_prof_gen_axis_state_start(axis=axis).decode()

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2002 0000 0000"
        )


def test_set_prof_gen_axis_state_start():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_state_start(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 2002 0000 {format_value(cargo2)}"
        )

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        cmd_string = TCU_TEST.set_prof_gen_axis_state_start(axis=axis, cargo2=cargo2).decode()

        assert is_tcu_write_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == format_value(axis)
        assert get_cmd_id(cmd_string) == M2MDCommandIdentifier.PROF_GEN_AXIS_STATE_START.value
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == format_value(cargo2)

        assert cmd_string.startswith(
            f"0320 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 2002 0000 {format_value(cargo2)}"
        )


def test_sw_rs_xx_sw_rise():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        for position in range(1, 21):
            expected_cmd_id = f"{M2MDCommandIdentifier.SW_RS_XX_SW_RISE.value[:2]}{hex(position)[2:].zfill(2)}"
            cmd_string = TCU_TEST.sw_rs_xx_sw_rise(axis=axis, position=position).decode()

            assert is_tcu_read_cmd(cmd_string)
            assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
            assert get_data_length(cmd_string) == str(DATA_LENGTH)
            assert get_address(cmd_string) == format_value(axis)
            assert get_cmd_id(cmd_string) == expected_cmd_id
            assert get_cargo1(cmd_string) == "0000"
            assert get_cargo2(cmd_string) == "0000"

            assert cmd_string.startswith(
                f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 30{hex(position)[2:].zfill(2)} 0000 0000"
            )

    for axis in range(1, NUM_M2MD_AXES + 1):
        for position in range(1, 21):
            expected_cmd_id = f"{M2MDCommandIdentifier.SW_RS_XX_SW_RISE.value[:2]}{hex(position)[2:].zfill(2)}"
            cmd_string = TCU_TEST.sw_rs_xx_sw_rise(axis=axis, position=position).decode()

            assert is_tcu_read_cmd(cmd_string)
            assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
            assert get_data_length(cmd_string) == str(DATA_LENGTH)
            assert get_address(cmd_string) == format_value(axis)
            assert get_cmd_id(cmd_string) == expected_cmd_id
            assert get_cargo1(cmd_string) == "0000"
            assert get_cargo2(cmd_string) == "0000"

            assert cmd_string.startswith(
                f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 30{hex(position)[2:].zfill(2)} 0000 0000"
            )


def test_sw_rs_xx_sw_fall():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        offset = 21
        for position in range(1, 21):
            expected_cmd_id = f"{M2MDCommandIdentifier.SW_RS_XX_SW_FALL.value[:2]}{hex(position + offset)[2:].zfill(2)}"
            cmd_string = TCU_TEST.sw_rs_xx_sw_fall(axis=axis, position=position).decode()

            assert is_tcu_read_cmd(cmd_string)
            assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
            assert get_data_length(cmd_string) == str(DATA_LENGTH)
            assert get_address(cmd_string) == format_value(axis)
            assert get_cmd_id(cmd_string) == expected_cmd_id
            assert get_cargo1(cmd_string) == "0000"
            assert get_cargo2(cmd_string) == "0000"

            assert cmd_string.startswith(
                f"0340 {get_expected_transaction_id_as_hex()} 0004 {axis.value} 30{hex(position + offset)[2:].zfill(2)} 0000 0000"
            )

    for axis in range(1, NUM_M2MD_AXES + 1):
        offset = 21
        for position in range(1, 21):
            expected_cmd_id = f"{M2MDCommandIdentifier.SW_RS_XX_SW_FALL.value[:2]}{hex(position + offset)[2:].zfill(2)}"
            cmd_string = TCU_TEST.sw_rs_xx_sw_fall(axis=axis, position=position).decode()

            assert is_tcu_read_cmd(cmd_string)
            assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
            assert get_data_length(cmd_string) == str(DATA_LENGTH)
            assert get_address(cmd_string) == format_value(axis)
            assert get_cmd_id(cmd_string) == expected_cmd_id
            assert get_cargo1(cmd_string) == "0000"
            assert get_cargo2(cmd_string) == "0000"

            assert cmd_string.startswith(
                f"0340 {get_expected_transaction_id_as_hex()} 0004 {format_value(axis)} 30{hex(position + offset)[2:].zfill(2)} 0000 0000"
            )


def test_tsm_latch():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.tsm_latch(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_LATCH.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 0000 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_get_tsm_current_value():
    cmd_string = TCU_TEST.get_tsm_current_value().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_CURRENT_VALUE.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 0001 0000 0000")


def test_set_tsm_current_value():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.set_tsm_current_value(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_CURRENT_VALUE.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 0001 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_get_tsm_current_offset():
    cmd_string = TCU_TEST.get_tsm_current_offset().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_CURRENT_OFFSET.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 0002 0000 0000")


def test_set_tsm_current_offset():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.set_tsm_current_offset(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_CURRENT_OFFSET.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 0002 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_tsm_adc_register_latch():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.tsm_adc_register_latch(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_REGISTER_LATCH.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1000 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_tsm_adc_id_register():
    cmd_string = TCU_TEST.tsm_adc_id_register().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_ID_REGISTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 1001 0000 0000")


def test_tsm_adc_configuration_register():
    cmd_string = TCU_TEST.tsm_adc_configuration_register().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_CONFIGURATION_REGISTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 1002 0000 0000")


def test_get_tsm_adc_hpf_register():
    cmd_string = TCU_TEST.get_tsm_adc_hpf_register().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_HPF_REGISTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 1003 0000 0000")


def test_set_tsm_adc_hpf_register():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.set_tsm_adc_hpf_register(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_HPF_REGISTER.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1003 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_get_tsm_adc_ofc_register():
    cmd_string = TCU_TEST.get_tsm_adc_ofc_register().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_OFC_REGISTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 1004 0000 0000")


def test_set_tsm_adc_ofc_register():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.set_tsm_adc_ofc_register(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_OFC_REGISTER.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1004 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_get_tsm_adc_fsc_register():
    cmd_string = TCU_TEST.get_tsm_adc_fsc_register().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_FSC_REGISTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 1006 0000 0000")


def test_set_tsm_adc_fsc_register():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.set_tsm_adc_fsc_register(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_FSC_REGISTER.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1006 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_tsm_adc_command_latch():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.tsm_adc_command_latch(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_COMMAND_LATCH.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1008 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_tsm_adc_command():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.tsm_adc_command(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_COMMAND.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 1009 {format_value(cargo1)} {format_value(cargo2)}"
    )


def test_tsm_adc_calibration():
    cargo1 = get_random_hex()
    cargo2 = get_random_hex()
    cmd_string = TCU_TEST.tsm_adc_calibration(cargo1=cargo1, cargo2=cargo2).decode()

    assert is_tcu_write_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ADC_CALIBRATION.value
    assert get_cargo1(cmd_string) == format_value(cargo1)
    assert get_cargo2(cmd_string) == format_value(cargo2)

    assert cmd_string.startswith(
        f"0320 {get_expected_transaction_id_as_hex()} 0004 0004 100A {format_value(cargo1)} {format_value(cargo2)}"
    )


# noinspection SpellCheckingInspection
def test_tsm_adc_value_xx_currentn():
    for sens_n in range(1, NUM_TSM_FRAMES * NUM_TSM_PROBES_PER_FRAME + 1):
        sens_a = 4 * sens_n
        cmd_string = TCU_TEST.tsm_adc_value_xx_currentn(probe=sens_n).decode()
        expected_cmd_id = "20" + hex(sens_a)[2:].zfill(2)

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.TSM.value
        assert get_cmd_id(cmd_string) == expected_cmd_id
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 20{hex(4 * sens_n)[2:].zfill(2)} 0000 0000"
        )


# noinspection SpellCheckingInspection
def test_tsm_adc_value_xx_biasn():
    for sens_n in range(1, NUM_TSM_FRAMES * NUM_TSM_PROBES_PER_FRAME + 1):
        sens_a = 4 * sens_n + 1
        cmd_string = TCU_TEST.tsm_adc_value_xx_biasn(probe=sens_n).decode()
        expected_cmd_id = "20" + hex(sens_a)[2:].zfill(2)

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.TSM.value
        assert get_cmd_id(cmd_string) == expected_cmd_id
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 20{hex(4 * sens_n + 1)[2:].zfill(2)} 0000 0000"
        )


# noinspection SpellCheckingInspection
def test_tsm_adc_value_xx_currentp():
    for sens_n in range(1, NUM_TSM_FRAMES * NUM_TSM_PROBES_PER_FRAME + 1):
        sens_a = 4 * sens_n + 2
        cmd_string = TCU_TEST.tsm_adc_value_xx_currentp(probe=sens_n).decode()
        expected_cmd_id = "20" + hex(sens_a)[2:].zfill(2)

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.TSM.value
        assert get_cmd_id(cmd_string) == expected_cmd_id
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 20{hex(4 * sens_n + 2)[2:].zfill(2)} 0000 0000"
        )


# noinspection SpellCheckingInspection
def test_tsm_adc_value_xx_biasp():
    for sens_n in range(1, NUM_TSM_FRAMES * NUM_TSM_PROBES_PER_FRAME + 1):
        sens_a = 4 * sens_n + 3
        cmd_string = TCU_TEST.tsm_adc_value_xx_biasp(probe=sens_n).decode()
        expected_cmd_id = "20" + hex(sens_a)[2:].zfill(2)

        assert is_tcu_read_cmd(cmd_string)
        assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
        assert get_data_length(cmd_string) == str(DATA_LENGTH)
        assert get_address(cmd_string) == CommandAddress.TSM.value
        assert get_cmd_id(cmd_string) == expected_cmd_id
        assert get_cargo1(cmd_string) == "0000"
        assert get_cargo2(cmd_string) == "0000"

        assert cmd_string.startswith(
            f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 20{hex(4 * sens_n + 3)[2:].zfill(2)} 0000 0000"
        )


def test_tsm_acq_counter():
    cmd_string = TCU_TEST.tsm_acq_counter().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.TSM.value
    assert get_cmd_id(cmd_string) == TSMCommandIdentifier.TSM_ACQ_COUNTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0004 20C0 0000 0000")


def test_vhk_psu_vmotor():
    cmd_string = TCU_TEST.vhk_psu_vmotor().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_PSU_VMOTOR.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0000 0000 0000")


def test_vhk_psu_vhi():
    cmd_string = TCU_TEST.vhk_psu_vhi().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_PSU_VHI.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0001 0000 0000")


def test_vhk_psu_vlow():
    cmd_string = TCU_TEST.vhk_psu_vlow().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_PSU_VLOW.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0002 0000 0000")


def test_vhk_psu_vmedp():
    cmd_string = TCU_TEST.vhk_psu_vmedp().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_PSU_VMEDP.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0003 0000 0000")


def test_vhk_psu_vmedn():
    cmd_string = TCU_TEST.vhk_psu_vmedn().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_PSU_VMEDN.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0004 0000 0000")


def test_ihk_psu_vmedn():
    cmd_string = TCU_TEST.ihk_psu_vmedn().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.IHK_PSU_VMEDN.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0005 0000 0000")


def test_ihk_psu_vmedp():
    cmd_string = TCU_TEST.ihk_psu_vmedp().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.IHK_PSU_VMEDP.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0006 0000 0000")


def test_ihk_psu_vlow():
    cmd_string = TCU_TEST.ihk_psu_vlow().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.IHK_PSU_VLOW.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0007 0000 0000")


def test_ihk_psu_vhi():
    cmd_string = TCU_TEST.ihk_psu_vhi().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.IHK_PSU_VHI.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0008 0000 0000")


def test_ihk_psu_vmotor():
    cmd_string = TCU_TEST.ihk_psu_vmotor().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.IHK_PSU_VMOTOR.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0009 0000 0000")


def test_thk_psu_first():
    cmd_string = TCU_TEST.thk_psu_first().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_PSU_FIRST.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000A 0000 0000")


def test_thk_m2md_first():
    cmd_string = TCU_TEST.thk_m2md_first().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_M2MD_FIRST.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000B 0000 0000")


def test_thk_psu_second():
    cmd_string = TCU_TEST.thk_psu_second().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_PSU_SECOND.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000C 0000 0000")


def test_thk_m2md_second():
    cmd_string = TCU_TEST.thk_m2md_second().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_M2MD_SECOND.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000D 0000 0000")


def test_thk_cts_q1():
    cmd_string = TCU_TEST.thk_cts_q1().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_Q1.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000E 0000 0000")


def test_thk_cts_q2():
    cmd_string = TCU_TEST.thk_cts_q2().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_Q2.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 000F 0000 0000")


def test_thk_cts_q3():
    cmd_string = TCU_TEST.thk_cts_q3().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_Q3.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0010 0000 0000")


def test_thk_cts_q4():
    cmd_string = TCU_TEST.thk_cts_q4().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_Q4.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0011 0000 0000")


def test_thk_cts_fpga():
    cmd_string = TCU_TEST.thk_cts_fpga().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_FPGA.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0012 0000 0000")


def test_thk_cts_ads1282():
    cmd_string = TCU_TEST.thk_cts_ads1282().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.THK_CTS_ADS1282.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0013 0000 0000")


def test_vhk_ths_ret():
    cmd_string = TCU_TEST.vhk_ths_ret().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.VHK_THS_RET.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0014 0000 0000")


def test_hk_acq_counter():
    cmd_string = TCU_TEST.hk_acq_counter().decode()

    assert is_tcu_read_cmd(cmd_string)
    assert get_transaction_id(cmd_string) == get_expected_transaction_id_as_hex()
    assert get_data_length(cmd_string) == str(DATA_LENGTH)
    assert get_address(cmd_string) == CommandAddress.HK.value
    assert get_cmd_id(cmd_string) == HKCommandIdentifier.HK_ACQ_COUNTER.value
    assert get_cargo1(cmd_string) == "0000"
    assert get_cargo2(cmd_string) == "0000"

    assert cmd_string.startswith(f"0340 {get_expected_transaction_id_as_hex()} 0004 0005 0015 0000 0000")


################
# Test simulator
################

TCU_SIMULATOR = TcuSimulator()


def test_tcu_firmware_id_sim():
    assert TCU_SIMULATOR.tcu_firmware_id() == "TCU Simulator"


def test_tcu_mode_sim():
    assert TCU_SIMULATOR.get_tcu_mode() == 0
    assert TCU_SIMULATOR.get_tcu_mode() == TcuMode.IDLE.value

    for tcu_mode in TcuMode:
        TCU_SIMULATOR.set_tcu_mode(tcu_mode)
        assert TCU_SIMULATOR.get_tcu_mode() == tcu_mode.value

    for tcu_mode in TcuMode:
        TCU_SIMULATOR.set_tcu_mode(tcu_mode.value)
        assert TCU_SIMULATOR.get_tcu_mode() == tcu_mode.value


def test_tcu_status_sim():
    # TODO
    TCU_SIMULATOR.tcu_status()


def test_tcu_simulated_sim():
    # TODO

    cargo2 = get_random_hex()
    TCU_SIMULATOR.tcu_simulated(cargo2)


def test_acq_curr_off_corr_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_acq_curr_off_corr(axis=axis, cargo2=cargo2)

        assert TCU_SIMULATOR.get_acq_curr_off_corr(axis=axis) == cargo2

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_acq_curr_off_corr(axis=axis, cargo2=cargo2)

        assert TCU_SIMULATOR.get_acq_curr_off_corr(axis=axis) == cargo2


def test_acq_curr_gain_corr_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_acq_curr_gain_corr(axis=axis, cargo2=cargo2)

        assert TCU_SIMULATOR.get_acq_curr_gain_corr(axis=axis) == cargo2

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_acq_curr_gain_corr(axis=axis, cargo2=cargo2)

        assert TCU_SIMULATOR.get_acq_curr_gain_corr(axis=axis) == cargo2


def test_acq_axis_a_curr_read_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        assert TCU_SIMULATOR.acq_axis_a_curr_read(axis=axis) == int(axis)

    for axis in range(1, NUM_M2MD_AXES + 1):
        assert TCU_SIMULATOR.acq_axis_a_curr_read(axis=axis) == axis


def test_acq_axis_b_curr_read_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        assert TCU_SIMULATOR.acq_axis_b_curr_read(axis=axis) == int(axis) + 3

    for axis in range(1, NUM_M2MD_AXES + 1):
        assert TCU_SIMULATOR.acq_axis_b_curr_read(axis=axis) == axis + 3


def test_prof_gen_axis_step_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_step(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_step(axis=axis) == cargo2

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_step(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_step(axis=axis) == cargo2


def test_prof_gen_axis_speed_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_speed(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_speed(axis=axis) == cargo2

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_speed(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_speed(axis=axis) == cargo2


def test_prof_gen_axis_state_start_sim():
    for axis in [CommandAddress.M2MD_1, CommandAddress.M2MD_2, CommandAddress.M2MD_3]:
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_state_start(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_state_start(axis=axis) == cargo2

    for axis in range(1, NUM_M2MD_AXES + 1):
        cargo2 = get_random_hex()
        TCU_SIMULATOR.set_prof_gen_axis_state_start(axis=axis, cargo2=cargo2)
        assert TCU_SIMULATOR.get_prof_gen_axis_state_start(axis=axis) == cargo2


def test_vhk_psu_vmotor_sim():
    assert TCU_SIMULATOR.vhk_psu_vmotor() == TcuSimulator.VHK_PSU_MOTOR


def test_vhk_psu_vhi_sim():
    assert TCU_SIMULATOR.vhk_psu_vhi() == TcuSimulator.VHK_PSU_VHI


def test_vhk_psu_vlow_sim():
    assert TCU_SIMULATOR.vhk_psu_vlow() == TcuSimulator.VHK_PSU_VLOW


def test_vhk_psu_vmedp_sim():
    assert TCU_SIMULATOR.vhk_psu_vmedp() == TcuSimulator.VHK_PSU_VMEDP


def test_vhk_psu_vmedn_sim():
    assert TCU_SIMULATOR.vhk_psu_vmedn() == TcuSimulator.VHK_PSU_VMEDN


def test_ihk_psu_vmedn_sim():
    assert TCU_SIMULATOR.ihk_psu_vmedn() == TcuSimulator.IHK_PSU_VMEDN


def test_ihk_psu_vmedp_sim():
    assert TCU_SIMULATOR.ihk_psu_vmedp() == TcuSimulator.IHK_PSU_VMEDP


def test_ihk_psu_vlow_sim():
    assert TCU_SIMULATOR.ihk_psu_vlow() == TcuSimulator.IHK_PSU_VLOW


def test_ihk_psu_vhi_sim():
    assert TCU_SIMULATOR.ihk_psu_vhi() == TcuSimulator.IHK_PSU_VHI


def test_ihk_psu_vmotor_sim():
    assert TCU_SIMULATOR.ihk_psu_vmotor() == TcuSimulator.IHK_PSU_VMOTOR


def test_thk_psu_first_sim():
    assert TCU_SIMULATOR.thk_psu_first() == TcuSimulator.THK_PSU_FIRST


def test_thk_m2md_first_sim():
    assert TCU_SIMULATOR.thk_m2md_first() == TcuSimulator.THK_M2MD_FIRST


def test_thk_psu_second_sim():
    assert TCU_SIMULATOR.thk_psu_second() == TcuSimulator.THK_PSU_SECOND


def test_thk_m2md_second_sim():
    assert TCU_SIMULATOR.thk_m2md_second() == TcuSimulator.THK_M2MD_SECOND


def test_thk_cts_q1_sim():
    assert TCU_SIMULATOR.thk_cts_q1() == TcuSimulator.THK_CTS_Q1


def test_thk_cts_q2_sim():
    assert TCU_SIMULATOR.thk_cts_q2() == TcuSimulator.THK_CTS_Q2


def test_thk_cts_q3_sim():
    assert TCU_SIMULATOR.thk_cts_q3() == TcuSimulator.THK_CTS_Q3


def test_thk_cts_q4_sim():
    assert TCU_SIMULATOR.thk_cts_q4() == TcuSimulator.THK_CTS_Q4


def test_thk_cts_fpga_sim():
    assert TCU_SIMULATOR.thk_cts_fpga() == TcuSimulator.THK_CTS_FPGA


def test_thk_cts_ads1282_sim():
    assert TCU_SIMULATOR.thk_cts_ads1282() == TcuSimulator.THK_CTS_ADS1282


def test_vhk_ths_ret_sim():
    assert TCU_SIMULATOR.vhk_ths_ret() == TcuSimulator.VHK_THS_RET


def test_hk_acq_counter_sim():
    assert TCU_SIMULATOR.hk_acq_counter() == TcuSimulator.HK_ACQ_COUNTER



import numpy as np
import math
import transforms3d as t3

from egse.coordinates.rotation_matrix import RotationMatrix


def affine_is_euclidian(matrix: np.ndarray) -> bool:
    """Checks if the given matrix is a pure solid-body Euclidian rotation + translation (no shear or scaling).

    We only need to check that:
        - The rotation part is orthogonal : R @ R.T = I
        - The det(R) = 1  (to check that the matrix does not represent a reflection)

    Args:
        matrix (np.ndarray): Matrix to check.

    Returns:
        True if the matrix is a pure solid-body Euclidian rotation + translation, False otherwise.
    """

    rotation = matrix[:3, :3]

    return np.allclose((rotation @ rotation.T), np.identity(3)) & np.allclose(np.linalg.det(matrix), 1)


def affine_inverse(matrix):
    """Returns the inverse of the given affine matrix.

    We assume that the given matrix is an affine transformation matrix that only involves rotation and translation,
    no zoom, no shear!  That is why we can invert it by simply transposing the rotation part and negating the
    translation part.

    Args:
        matrix (np.ndarray): Augmented matrix to invert.

    Returns:
        Inverted affine matrix.
    """

    if affine_is_euclidian(matrix):
        # Separate the given augmented matrix into rotation and translation

        rotation = matrix[:3, :3]
        translation = matrix[:3, 3]

        # Invert the rotation and the translation

        inverse_rotation = rotation.T
        inverse_translation = -translation

        # The inverse affine matrix is composed of
        #   - Rotation: R^-1
        #   - Translation: -(R^-1 . t)

        result = np.identity(4)
        result[:3, :3] = inverse_rotation
        result[:3, 3] = np.dot(inverse_rotation, inverse_translation)

        return result

    else:
        print("WARNING: This is not a rigid-body transformation matrix")
        # print(f"R.T-based  (.6f) = \n {np.round(result,6)}")
        # print(f"np.inverse (.6f) = \n {np.round(np.linalg.inv(matrix),6)}")
        return np.linalg.inv(matrix)


def affine_matrix_from_points(
    v0: np.ndarray, v1: np.ndarray, shear: bool = False, scale: bool = False, use_svd: bool = True
) -> np.ndarray:
    """Computes the homogeneous affine transform matrix that best maps one set of points to another.

    Returns affine transform matrix to register two point sets.

    v0 and v1 are shape (ndims, \*) arrays of at least ndims non-homogeneous
    coordinates, where ndims is the dimensionality of the coordinate space.

    If shear is False, a similarity transformation matrix is returned.
    If also scale is False, a rigid/Euclidean transformation matrix
    is returned.

    By default, the algorithm by Hartley and Zissermann [15] is used.
    If usesvd is True, similarity and Euclidean transformation matrices
    are calculated by minimizing the weighted sum of squared deviations
    (RMSD) according to the algorithm by Kabsch [8].
    Otherwise, and if ndims is 3, the quaternion based algorithm by Horn [9]
    is used, which is slower when using this Python implementation.

    The returned matrix performs rotation, translation and uniform scaling
    (if specified).

    >>> v0 = [[0, 1031, 1031, 0], [0, 0, 1600, 1600]]
    >>> v1 = [[675, 826, 826, 677], [55, 52, 281, 277]]
    >>> affine_matrix_from_points(v0, v1)
    array([[   0.14549,    0.00062,  675.50008],
           [   0.00048,    0.14094,   53.24971],
           [   0.     ,    0.     ,    1.     ]])
    >>> T = translation_matrix(np.random.random(3)-0.5)
    >>> R = random_rotation_matrix(np.random.random(3))
    >>> S = scale_matrix(random.random())
    >>> M = concatenate_matrices(T, R, S)
    >>> v0 = (np.random.rand(4, 100) - 0.5) * 20
    >>> v0[3] = 1
    >>> v1 = np.dot(M, v0)
    >>> v0[:3] += np.random.normal(0, 1e-8, 300).reshape(3, -1)
    >>> M = affine_matrix_from_points(v0[:3], v1[:3])
    >>> np.allclose(v1, np.dot(M, v0))
    True

    More examples in superimposition_matrix()

    References: This function was extracted from the original transformations.py written by Christoph Golke:
                https://www.lfd.uci.edu/~gohlke/code/transformations.py.html

    Args:
        v0 (np.ndarray): Set of points to transform.
        v1 (np.ndarray): Set of points to transform to.
        shear (bool): Indicates whether a full affine transform is allowed (i.e. incl. shear).
        scale (bool): Indicates whether a full affine transform is allowed (i.e. incl. zoom).
        use_svd (bool): Indicates whether to use SVD-base solution (Singular Value Decomposition) for
                        rigid/similarity transformations.  If False and ndims = 3, the quaternion-based solution is used.

    Returns:
        Best affine transformation matrix to map `v0` to `v1`.
    """

    v0 = np.array(v0, dtype=np.float64, copy=True)
    v1 = np.array(v1, dtype=np.float64, copy=True)

    num_dimensions = v0.shape[0]
    if num_dimensions < 2 or v0.shape[1] < num_dimensions or v0.shape != v1.shape:
        print(
            f"num_dimensions {num_dimensions} v0/1.shape {v0.shape} {v1.shape} v0/1 class {v0.__class__} {v1.__class__}"
        )
        raise ValueError("input arrays are of wrong shape or type")

    # First set of coordinates

    t0 = -np.mean(v0, axis=1)  # Move centroids to origin
    v0 += t0.reshape(num_dimensions, 1)
    matrix_0 = np.identity(num_dimensions + 1)
    matrix_0[:num_dimensions, num_dimensions] = t0  # (I | t0)

    # Second set of coordinates

    t1 = -np.mean(v1, axis=1)  # Move centroids to origin
    v1 += t1.reshape(num_dimensions, 1)
    matrix_1 = np.identity(num_dimensions + 1)
    matrix_1[:num_dimensions, num_dimensions] = t1  # (I | t1)

    if shear:
        # Affine transformation
        A = np.concatenate((v0, v1), axis=0)
        svd_u, svd_s, svd_vh = np.linalg.svd(A.T)  # Singular Value Decomposition -> U, S, Vh
        svd_vh = svd_vh[:num_dimensions].T
        B = svd_vh[:num_dimensions]
        C = svd_vh[num_dimensions : 2 * num_dimensions]
        t = np.dot(C, np.linalg.pinv(B))
        t = np.concatenate((t, np.zeros((num_dimensions, 1))), axis=1)
        M = np.vstack((t, ((0.0,) * num_dimensions) + (1.0,)))

    elif use_svd or num_dimensions != 3:
        # Rigid transformation via SVD of covariance matrix
        svd_u, svd_s, svd_vh = np.linalg.svd(np.dot(v1, v0.T))
        # rotation matrix from SVD orthonormal bases
        R = np.dot(svd_u, svd_vh)
        if np.linalg.det(R) < 0.0:
            # R does not constitute right-handed system
            R -= np.outer(svd_u[:, num_dimensions - 1], svd_vh[num_dimensions - 1, :] * 2.0)
            svd_s[-1] *= -1.0
        # homogeneous transformation matrix
        M = np.identity(num_dimensions + 1)
        M[:num_dimensions, :num_dimensions] = R

    else:
        # Rigid transformation matrix via quaternion
        # compute symmetric matrix N
        xx, yy, zz = np.sum(v0 * v1, axis=1)
        xy, yz, zx = np.sum(v0 * np.roll(v1, -1, axis=0), axis=1)
        xz, yx, zy = np.sum(v0 * np.roll(v1, -2, axis=0), axis=1)
        N = [
            [xx + yy + zz, 0.0, 0.0, 0.0],
            [yz - zy, xx - yy - zz, 0.0, 0.0],
            [zx - xz, xy + yx, yy - xx - zz, 0.0],
            [xy - yx, zx + xz, yz + zy, zz - xx - yy],
        ]
        # quaternion: eigenvector corresponding to most positive eigenvalue
        eigenvalues, eigenvectors = np.linalg.eigh(N)
        quaternion = eigenvectors[:, np.argmax(eigenvalues)]
        quaternion /= _vector_norm(quaternion)  # Normalised quaternion
        # Quaternion -> Homogeneous rotation matrix
        M = quaternion_matrix(quaternion)

    if scale and not shear:
        # Affine transformation; scale is ratio of RMS deviations from centroid
        v0 *= v0
        v1 *= v1
        M[:num_dimensions, :num_dimensions] *= math.sqrt(np.sum(v1) / np.sum(v0))

    # Move centroids back
    M = np.dot(np.linalg.inv(matrix_1), np.dot(M, matrix_0))
    M /= M[num_dimensions, num_dimensions]

    return M


def _vector_norm(data: np.ndarray, axis: str | None = None, out=None):
    """Returns the length, i.e. Euclidean norm, of the given array along the given axis.

    >>> v = np.random.random(3)
    >>> n = vector_norm(v)
    >>> np.allclose(n, np.linalg.norm(v))
    True
    >>> v = np.random.rand(6, 5, 3)
    >>> n = vector_norm(v, axis=-1)
    >>> np.allclose(n, np.sqrt(np.sum(v*v, axis=2)))
    True
    >>> n = vector_norm(v, axis=1)
    >>> np.allclose(n, np.sqrt(np.sum(v*v, axis=1)))
    True
    >>> v = np.random.rand(5, 4, 3)
    >>> n = np.empty((5, 3))
    >>> vector_norm(v, axis=1, out=n)
    >>> np.allclose(n, np.sqrt(np.sum(v*v, axis=1)))
    True
    >>> vector_norm([])
    0.0
    >>> vector_norm([1])
    1.0

    This function is called by affine_matrix_from_points when usesvd=False

    References: This function was extracted from the original transformations.py written by Christoph Golke:
                https://www.lfd.uci.edu/~gohlke/code/transformations.py.html

    """

    data = np.array(data, dtype=np.float64, copy=True)

    if out is None:
        if data.ndim == 1:
            return math.sqrt(np.dot(data, data))
        data *= data
        out = np.atleast_1d(np.sum(data, axis=axis))
        np.sqrt(out, out)

        return out
    else:
        data *= data
        np.sum(data, axis=axis, out=out)
        return np.sqrt(out, out)


def quaternion_matrix(quaternion: np.ndarray):
    """Returns the homogeneous rotation matrix from the given quaternion.

    >>> M = quaternion_matrix([0.99810947, 0.06146124, 0, 0])
    >>> np.allclose(M, rotation_matrix(0.123, [1, 0, 0]))
    True
    >>> M = quaternion_matrix([1, 0, 0, 0])
    >>> np.allclose(M, np.identity(4))
    True
    >>> M = quaternion_matrix([0, 1, 0, 0])
    >>> np.allclose(M, np.diag([1, -1, -1, 1]))
    True

    This function is called by affine_matrix_from_points when usesvd=False

    References: This function was extracted from the original transformations.py written by Christoph Golke:
                https://www.lfd.uci.edu/~gohlke/code/transformations.py.html

    Args:
        quaternion (np.ndarray): Quaternion to convert to a rotation matrix.

    Returns:
        Homogeneous rotation matrix corresponding to the given quaternion.
    """

    _EPS = np.finfo(float).eps * 5

    q = np.array(quaternion, dtype=np.float64, copy=True)
    n = np.dot(q, q)

    if n < _EPS:
        return np.identity(4)

    q *= math.sqrt(2.0 / n)
    q = np.outer(q, q)
    return np.array(
        [
            [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0], 0.0],
            [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0], 0.0],
            [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2], 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )


def rigid_transform_3d(dataset_a: np.ndarray, dataset_b: np.ndarray):
    """Returns best translation and rotation to align points in dataset A to points in dataset B.

    Args:
        dataset_a (np.ndarray): First 3xn dataset of points (dataset A).
        dataset_b (np.ndarray): Second 3xn dataset of points (dataset B).

    References: Nghia Ho - 2013 - http://nghiaho.com/?page_id=671
                "Finding optimal rotation and translation between corresponding 3D points"
                Based on "A Method for Registration of 3-D Shapes", by Besl and McKay, 1992.

    This is based on Singular Value Decomposition (SVD)
    -> It is equivalent to affine_matrix_from_points with parameter use_svd=True

    Returns:
        Transformation matrix to align points in dataset A to points in dataset B.
    """

    dataset_a_transposed = dataset_a.T
    dataset_b_transposed = dataset_b.T

    assert len(dataset_a_transposed) == len(dataset_b_transposed)

    num_points = dataset_a_transposed.shape[0]  # Total points

    centroid_a = np.mean(dataset_a_transposed, axis=0)
    centroid_b = np.mean(dataset_b_transposed, axis=0)

    # Centre the points
    a_centered = dataset_a_transposed - np.tile(centroid_a, (num_points, 1))
    b_centered = dataset_b_transposed - np.tile(centroid_b, (num_points, 1))

    # @ is matrix multiplication for array
    covariance_matrix = np.transpose(a_centered) @ b_centered  # Covariance matrix H

    svd_u, svd_s, svd_vh = np.linalg.svd(covariance_matrix)  # SVD(H) = [U, S, V]

    rotation = svd_vh.T @ svd_u.T  # Rotation matrix R

    # Special reflection case
    # There’s a special case when finding the rotation matrix that you have to take care of. Sometimes the SVD will
    # return a "reflection" matrix, which is numerically correct but is actually nonsense in real life. This is
    # addressed by checking the determinant of R (from SVD above) and seeing if it’s negative (-1). If it is then the
    # 3rd column of V is multiplied by -1.

    if np.linalg.det(rotation) < 0:
        print("Reflection detected")
        svd_vh[2, :] *= -1
        rotation = svd_vh.T @ svd_u.T

    translation = -rotation @ centroid_a.T + centroid_b.T

    result = np.identity(4)
    result[:3, :3] = rotation
    result[:3, 3] = translation

    return result


def translation_rotation_to_transformation(
    translation: np.ndarray,
    rotation: np.ndarray,
    rotation_config: str = "sxyz",
    active: bool = True,
    degrees: bool = True,
    translation_first: bool = False,
):
    """Returns the transformation matrix from the given translation and rotation.

    Args:
        translation (np.ndarray): Translation vector.
        rotation (np.ndarray): Rotation vector.
        rotation_config (str): Order in which the rotation about the three axes are chained.
        active (bool): Indicates if the rotation is active (object rotates IN a fixed coordinate system) or passive
                           (coordinate system rotates AROUND a fixed object).  Even if two angles are zero, the match
                           between angle orders and rot_config is still critical
        degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
        translation_first (bool): Indicates the order of the translation and rotation in the transformation matrix.
                                  False if the first three rows of the transformation matrix are (R t). This is the
                                  usual convention.
                                  True if the first three rows of the transformation matrix are (R Rt).  This is used
                                  in the hexapod.

    Returns:
        Transformation matrix corresponding to the given translation and rotation.
    """

    zoom = np.array([1, 1, 1])
    shear = np.array([0, 0, 0])
    translation = np.array(translation)
    if degrees:
        rotation = np.array([np.deg2rad(item) for item in rotation])
    rotation_x, rotation_y, rotation_z = rotation
    rotation_matrix = RotationMatrix(rotation_x, rotation_y, rotation_z, rotation_config=rotation_config, active=active)

    if translation_first:
        result = np.identity(4)
        result[:3, :3] = rotation_matrix.rotation_matrix
        result[:3, 3] = rotation_matrix.rotation_matrix @ translation
    else:
        result = t3.affines.compose(translation, rotation_matrix.rotation_matrix, Z=zoom, S=shear)

    return result


def translation_rotation_from_transformation(
    transformation: np.ndarray,
    rotation_config: str = "sxyz",
    active: bool = True,
    degrees: bool = True,
    translation_first: bool = False,
):
    """Extracts the translation and rotation vector from the given transformation matrix.

    Args:
    transformation (np.ndarray): Transformation matrix.
    rotation_config (str): Order in which the rotation about the three axes are chained.
    active (bool): Indicates if the rotation is active (object rotates IN a fixed coordinate system) or passive
                       (coordinate system rotates AROUND a fixed object).  Even if two angles are zero, the match
                       between angle orders and rot_config is still critical
    degrees (bool): Indicates whether the rotation angles are specified in degrees, rather than radians.
    translation_first (bool): Indicates the order of the translation and rotation in the transformation matrix.
                              False if the first three rows of the transformation matrix are (R t). This is the
                              usual convention.
                              True if the first three rows of the transformation matrix are (R Rt).  This is used
                              in the hexapod.

    Returns:
        Translation and rotation vector for the given transformation matrix.
    """

    translation = transformation[:3, 3]
    rotation = t3.euler.mat2euler(transformation, axes=rotation_config)
    if degrees:
        rotation = np.array([np.rad2deg(item) for item in rotation])
    if translation_first:
        translation = transformation[:3, :3].T @ translation

    return translation, rotation


tr2t = translation_rotation_to_transformation
t2tr = translation_rotation_from_transformation


def vector_plane_intersection(vector, frame, epsilon=1.0e-6):
    """Returns the coordinates of the insection of a vector with a plane.

    The origin of the input vector is:
        vector.reference_frame.get_origin().coordinates[:3]

    The direction of the input vector is:
        vector.coordinates[:3]

    In all cases, the coordinates of the intersection point are provided as a Point object, in "frame" coordinates

    Args:
        vector (Vector): Input vector.
        frame (ReferenceFrame): Reference frame for which the xy-plane is the target plane for intersection.
        epsilon (float):


    References:
        https://stackoverflow.com/questions/5666222/3d-line-plane-intersection
    """

    from egse.coordinates.point import Point

    if vector.reference_frame == frame:
        # The point is defined in frame => the origin of the vector is the origin of the target plane.
        return np.array([0, 0, 0])
    else:
        # Express all inputs in the given reference frame

        vector_origin = Point(
            vector.reference_frame.get_origin().coordinates[:3], reference_frame=vector.reference_frame, name="ptorig"
        ).express_in(frame)[:3]  # Vector Origin (p0)
        vector_end = vector.express_in(frame)[:3]  # Vector end (p1)

        vector_u = vector_end - vector_origin  # Vector (u)

        plane_origin = frame.get_origin().coordinates[:3]  # Origin of the reference frame (p_co)
        plane_normal = frame.get_axis("z").coordinates[:3]  # Normal to the plane (p_no)

        # Vector to normal 'angle'
        dot = np.dot(vector_u, plane_normal)  # dot = p_no * u

        # Test if there is an intersection (and if it's unique)
        # -> input vector and normal mustn't be perpendicular, else the vector is // to the plane or inside it

        if np.allclose(dot, 0.0, atol=epsilon):
            print("The input vector is // to the plane normal (or inside the plane)")
            print("-> There exists no intersection (or an infinity of them)")
            return None
        else:
            # Vector from the point in the plane to the origin of the vector (w)
            plane_to_vector = vector_origin - plane_origin  # w = p0 - p_co

            # Solution  ("how many 'vectors' away is the intersection ?")
            factor = -np.dot(plane_normal, plane_to_vector) / dot  # fac = -(plane * w) / fac

            return Point(vector_origin + (vector_u * factor), reference_frame=frame)



import asyncio

import pytest

from egse.log import logger
from egse.registry.service import ZMQMicroservice
from egse.system import get_host_ip


@pytest.mark.asyncio
async def test_zmq_microservice_initialization():
    print()

    service = ZMQMicroservice("vanilla", "plain")

    assert service.service_name == "vanilla"
    assert service.service_type == "plain"
    assert service.host_ip == get_host_ip()

    print(f"{service.rep_port = }")

    assert service.rep_port != 0

    await service._cleanup()


# Registration to the service registry will time out in 5s
# @pytest.mark.timeout(10)
@pytest.mark.asyncio
async def test_zmq_microservices_registration(registry_service):
    service = ZMQMicroservice("vanilla", "plain")

    service_task = asyncio.create_task(service.start())

    await asyncio.sleep(5.0)

    await service.stop()
    await service_task

    logger.info("Service task completed")

    service_task.cancel()
    try:
        await service_task
    except asyncio.CancelledError:
        logger.info("Service task was cancelled as expected.")



import logging

from egse.log import logger


def test_log_levels(caplog):
    orig_level = logger.level

    # The egse logger doesn't propagate messages to parent loggers, so we
    # have to add the caplog handler in order to capture logging messages for this test.
    logger.addHandler(caplog.handler)
    logger.setLevel(logging.INFO)

    level = logger.getEffectiveLevel()
    assert level == logging.INFO

    caplog.clear()

    try:
        for name, level_ in (
            ("DEBUG", logging.DEBUG),
            ("INFO", logging.INFO),
            ("WARNING", logging.WARNING),
        ):
            logger.log(level_, f"{name} logging message")

        if level <= logging.DEBUG:
            assert "DEBUG logging message" in caplog.text
        else:
            assert "DEBUG logging message" not in caplog.text

        if level <= logging.INFO:
            assert "INFO logging message" in caplog.text
        else:
            assert "INFO logging message" not in caplog.text

        if level <= logging.WARNING:
            assert "WARNING logging message" in caplog.text
        else:
            assert "WARNING logging message" not in caplog.text

    finally:
        logger.removeHandler(caplog.handler)
        logger.setLevel(orig_level)

    caplog.clear()



# An example plugin for the `cgse {start,stop,status} service` command from `cgse-core`.
#
import subprocess
import sys

import rich
import typer

from egse.system import redirect_output_to_log

daq6510 = typer.Typer(name="daq6510", help="DAQ6510 Data Acquisition Unit, Keithley, temperature monitoring")


@daq6510.command(name="start")
def start_daq6510():
    """Start the daq6510 service."""
    rich.print("Starting service daq6510")

    out = redirect_output_to_log("daq6510_cs.start.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_cs", "start"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@daq6510.command(name="stop")
def stop_daq6510():
    """Stop the daq6510 service."""
    rich.print("Terminating service daq6510")

    out = redirect_output_to_log("daq6510_cs.stop.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_cs", "stop"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@daq6510.command(name="status")
def status_daq6510():
    """Print status information on the daq6510 service."""

    proc = subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_cs", "status"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
    )

    stdout, stderr = proc.communicate()

    rich.print(stdout.decode(), end="")
    # if stderr:
    #     rich.print(f"[red]{stderr.decode()}[/]")


@daq6510.command(name="start-sim")
def start_daq6510_sim():
    """Start the DAQ6510 Simulator."""
    rich.print("Starting service DAQ6510 Simulator")

    out = redirect_output_to_log("daq6510_sim.start.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_sim", "start"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@daq6510.command(name="stop-sim")
def stop_daq6510_sim():
    """Stop the DAQ6510 Simulator."""
    rich.print("Terminating the DAQ6510 simulator.")

    out = redirect_output_to_log("daq6510_sim.stop.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_sim", "stop"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@daq6510.command(name="start-mon")
def start_daq6510_mon():
    """Start the daq6510 monitoring service."""
    rich.print("Starting monitoring service daq6510_mon")

    out = redirect_output_to_log("daq6510_mon.start.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_mon", "start"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@daq6510.command(name="stop-mon")
def stop_daq6510_mon():
    """Stop the daq6510 monitoring service."""
    rich.print("Stopping monitoring service daq6510_mon")

    out = redirect_output_to_log("daq6510_mon.stop.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.tempcontrol.keithley.daq6510_mon", "stop"],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


if __name__ == "__main__":
    daq6510()



__all__ = [
    "show_processes",
]

import re

from egse.process import ProcessInfo, get_processes


def show_processes():
    def filter_procs(pi: ProcessInfo):
        pattern = r"tcu_(ui|cs)"

        return re.search(pattern, pi.command)

    return get_processes(filter_procs)
